/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

/* Algorithme de la boulangerie de Lamport

   assure une section critique entre n threads */

#include "prelude.h"
#include <pthread.h>

// le nombre de threads
int n;

// Le qualificatif `volatile` ci-dessous limite les optimisations
// que le compilateur C ferait en considérant qu'il s'agit d'un code
// séquentiel. Une alternative est de compiler avec -O0 seulement.
volatile int *number;
volatile bool *entering;

// comparaison lexicographique (number[j], j) < (number[i], i)
bool smaller(int j, int i) {
  return number[j] < number[i] || (number[j] == number[i] && j < i);
}

void lock(int i) {
  entering[i] = true;
  for (int j = 0; j < n; j++)
    if (number[j] > number[i])
      number[i] = number[j];
  number[i]++; // 1 + max(number[0],...,number[n-1])
  entering[i] = false;
  for (int j = 0; j < n; j++) {
    while (entering[j]); // attente active
    while (number[j] != 0 && smaller(j, i)); // attente active
  }
}

void unlock(int i) {
  number[i] = 0;
}

void *create(void *arg) {
  int i = *((int *)arg);
  lock(i);
  // Ici on fait quelque chose d'un peu long, histoire que les affichages
  // se mélangent si la section critique n'est pas assurée. Il est facile de le
  // constater en remplaçant la fonction `lock` ci-dessus par une
  // fonction sans effet.
  printf("thread %d\n", i);
  for (int j = 0; j < 30; j++)
    printf(" %d", i);
  printf("\n");
  unlock(i);
  return NULL;
}


int main(int argc, char **argv) {
  n = argc == 1 ? 5 : atoi(argv[1]);
  number = calloc(n, sizeof(int));
  for (int i = 0; i < n; i++) number[i] = 0;
  entering = calloc(n, sizeof(bool));
  for (int i = 0; i < n; i++) entering[i] = false;
  int *pid = calloc(n, sizeof(int));
  pthread_t *thread = malloc(n * sizeof(pthread_t));
  for (int i = 0; i < n; i++) {
    pid[i] = i;
    pthread_create(&thread[i], NULL, create, &pid[i]);
  }
  for (int i = 0; i < n; i++) {
    pthread_join(thread[i], NULL);
  }
}
